#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# Inspired by https://github.com/silvester747/aio_marantz_avr
#
# This is to control Denon/Marantz AVR devices
#
# Copyright (c) 2020 François Wautier
#
# Note large part of this code was taken from scapy and other opensource software
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE

"""Control of an AVR over Telnet."""

import asyncio
import logging
import re
from enum import Enum
from typing import Any, List, Mapping, Optional, Callable
from .enums import InputSource, Power, SurroundMode, ChannelBias, EcoMode, AudioInput

#Some replacement for the surrond sound format
SSTRANSFORM = [("Audio-"," "),("Dd","Dolby Digital "),("DD","Dolby Digital "),("Dts","DTS"),['Mstr','Master '],
                     ("Dsur","Digital Surround "),("Mtrx","Matrix"),("Dscrt","Discrete "),(" Es "," ES ")]
EXTRAS = ["SSINFAI"]

def cc_string(identifier: str) -> str:
    """ From https://stackoverflow.com/questions/29916065/how-to-do-camelcase-split-in-python """
    matches = re.finditer('.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)', identifier)
    return " ".join([m.group(0) for m in matches])

def only_int(val: str) -> str:
    return "".join([x for x in val if x in ['0','1','2','3','4','5','6','7','8','9']])


class AvrError(Exception):
    """Base class for all errors returned from an AVR."""

    pass

class AvrTimeoutError(AvrError):
    """A request to the AVR has timed out."""

    pass


async def avr_factory(name: str, host: str, port: int = 23, timeout: float = 3.0) -> "MDAVR":
    """Connect to an AVR.

        :param name: The name of this device.
        :type url: str
        :param addr: The device IP address
        :type name: str
        :returns: A device instance or None if connection cannot be established
        :rtype: MDAVR
    """
    try:
        reader, writer = await asyncio.open_connection(host, port=port)
        return MDAVR(name, reader, writer, timeout)
    except:
        return None


def _on_off_from_bool(value: bool) -> str:
    if value:
        return "ON"
    else:
        return "OFF"


def _on_off_to_bool(value: str) -> bool:
    return value == "ON"


class _CommandDef:
    code: str
    label: str
    vals: Optional[Enum]

    def __init__(self,  label: str, vals: Any):
        self.label = label
        self.values = vals


class MDAVR:
    """Connection to a Marantz AVR over Telnet.

    Uses `connect` to create a connection to the AVR.
    """

    CMDS_DEFS: Mapping[str,_CommandDef] = {
        "PW": _CommandDef("Power", Power),
        "MU": _CommandDef("Mute", None),
        "MV": _CommandDef("Volume", None),
        "SI": _CommandDef("Source", InputSource),
        "MS": _CommandDef("Surround Mode", SurroundMode),
        "CV": _CommandDef("Channel Bias", ChannelBias),
        "ECO": _CommandDef("Eco Mode", EcoMode),
    }

    _reader: asyncio.StreamReader
    _writer: asyncio.StreamWriter
    _timeout: float

    def __init__(
        self, name: str, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, timeout: float,
    ):
        self.name = name
        self._reader = reader
        self._writer = writer
        self._timeout = timeout
        self.status = {}
        self.maxvol = 98      #Good default ;)
        self.alive = True
        self.queue = []
        for x in self.CMDS_DEFS:
            self.status[self.CMDS_DEFS[x].label]="-"
        self.cvend = True
        self.notify = None
        #Start reading
        self.rtask = asyncio.get_event_loop().create_task(self._do_read())

    def _get_current(self,cmd):
        return self.status[self.CMDS_DEFS[cmd].label]

    def _get_list(self,cmd):
        return [cc_string(x.name) for x in list(self.CMDS_DEFS[cmd].values)]

    #API Starts here
    @property
    def power(self) -> Optional[Power]:
        """Power state of the AVR."""
        return self._get_current("PW")

    @property
    def muted(self) -> Optional[bool]:
        """Boolean if volume is currently muted."""
        return self._get_current("MU")

    @property
    def volume(self) -> Optional[float]:
        """Volume level of the AVR zone (00..max_volume)."""
        return self._get_current("MV")

    @property
    def max_volume(self) -> Optional[float]:
        """Maximum volume level of the AVR zone."""
        return self.maxvol

    @property
    def source(self) -> str:
        """Name of the current input source."""
        return self._get_current("SI")

    @property
    def source_list(self) -> List[str]:
        """List of available input sources."""
        return self._get_list("SI")

    @property
    def sound_mode(self) -> str:
        """Name of the current sound mode."""
        return self._get_current("MS")

    @property
    def sound_mode_list(self) -> List[str]:
        """List of available sound modes."""
        return self._get_list("MS")

    @property
    def eco_mode(self) -> str:
        """Current ECO mode."""
        return self._get_current("ECO")

    @property
    def eco_mode_list(self) -> List[str]:
        """List of available exo modes."""
        return self._get_list("ECO")

    @property
    def channels_bias(self) -> Mapping[str,float]:
        return self._get_current("CV")

    @property
    def channels_bias_list(self) -> List[str]:
        """List of currently available."""
        return [x for x in self._get_current("CV").keys()]


    async def refresh(self) -> None:
        """Refresh all properties from the AVR."""

        for cmd_def in self.CMDS_DEFS:
            fut = await self._send_command(cmd_def,"?")
            try:
                await self._wait_for_response_with_timeout(fut)
            except asyncio.TimeoutError:
                self._clean_fut(fut)
                raise AvrTimeoutError

    async def turn_on(self) -> None:
        """Turn the AVR on."""
        fut = await self._send_command("PW", "ON")
        try:
            await self._wait_for_response_with_timeout(fut)
        except asyncio.TimeoutError:
            self._clean_fut(fut)
            raise AvrTimeoutError

    async def turn_off(self) -> None:
        """Turn the AVR off."""
        fut = await self._send_command("PW", "STANDBY")
        try:
            await self._wait_for_response_with_timeout(fut)
        except asyncio.TimeoutError:
            self._clean_fut(fut)
            raise AvrTimeoutError


    async def mute_volume(self, mute: bool) -> None:
        """Mute or unmute the volume.

        Arguments:
        mute -- True to mute, False to unmute.
        """
        fut = await self._send_command("MU", _on_off_from_bool(mute))
        try:
            await self._wait_for_response_with_timeout(fut)
        except asyncio.TimeoutError:
            self._clean_fut(fut)
            raise AvrTimeoutError

    async def set_volume(self, level: float) -> None:
        """Set the volume level.

        Arguments:
        level -- An integer value between 0 and `max_volume`.
        """
        if level > self.maxvol:
            level = maxvol
        if int(10*level)%10:
            #Needs to be a nultiple of 5
            level = int(5*round(10*level/5))
        else:
            level = int(level)
        fut = await self._send_command("MV",f"{level:02}")
        try:
            await self._wait_for_response_with_timeout(fut)
        except asyncio.TimeoutError:
            self._clean_fut(fut)
            raise AvrTimeoutError

    async def volume_up(self) -> None:
        """Turn the volume level up one notch."""
        fut = await self._send_command("MV","UP")
        try:
            await self._wait_for_response_with_timeout(fut)
        except asyncio.TimeoutError:
            self._clean_fut(fut)
            raise AvrTimeoutError

    async def volume_down(self) -> None:
        """Turn the volume level down one notch."""
        fut = await self._send_command("MV","DOWN")
        try:
            await self._wait_for_response_with_timeout(fut)
        except asyncio.TimeoutError:
            self._clean_fut(fut)
            raise AvrTimeoutError

    async def set_channel_bias(self, chan: str, level: float) -> None:
        """Set the volume level.

        Arguments:
        chan  -- channel to set
        level -- A float value between -12.0 and +12.0
        """
        if chan not in self.channels_bias:
            logging.warning(f"Channel {chan} is not available right now.")
            await asyncio.sleep(0)
            return


        if self.channels_bias[chan] != level:
            chan = chan.replace(" ","")
            level = level+50 #50 is 0dB
            if level < 38:
                level = 38
            elif level > 62:
                level = 62
            if int(10*level)%10:
                #Needs to be a nultiple of 5
                level = int(5*round(10*level/5))
            else:
                level = int(level)

            cmd = None
            for x in self.CMDS_DEFS["CV"].values:
                if x.name == chan:
                    cmd = x.value
                    break
            if cmd:
                fut = await self._send_command("CV", f"{cmd} {level:02}")
                try:
                    await self._wait_for_response_with_timeout(fut)
                except asyncio.TimeoutError:
                    self._clean_fut(fut)
                    raise AvrTimeoutError
            else:
                logging.error(f"Channel {chan} should exist. This should not have happened.")

    async def channel_bias_up(self, chan: str) -> None:
        """Turn the volume level up one notch."""
        if chan not in self.channels_bias:
            logging.warning(f"Channel {chan} is not available right now.")
            await asyncio.sleep(0)
            return
        if self.channels_bias[chan] == 12:
            #We are at the limit. It won't respond
            logging.debugf(f"Channel {chan} it at the upper limit.")
            await asyncio.sleep(0)
            return

        chan = chan.replace(" ","")
        cmd = None
        for x in self.CMDS_DEFS["CV"].values:
            if x.name == chan:
                cmd = x.value
                break
        if cmd:
            fut = await self._send_command("CV", f"cmd UP")
            try:
                await self._wait_for_response_with_timeout(fut)
            except asyncio.TimeoutError:
                self._clean_fut(fut)
                raise AvrTimeoutError
        else:
            logging.error(f"Channel {chan} should exist. This should not have happened.")

    async def channel_bias_down(self, chan: str) -> None:
        """Turn the volume level down one notch."""
        if chan not in self.channels_bias:
            logging.warning(f"Channel {chan} is not available right now.")
            await asyncio.sleep(0)
            return
        if self.channels_bias[chan] == -12:
            #We are at the limit. It won't respond
            logging.debugf(f"Channel {chan} it at the lowewr limit.")
            await asyncio.sleep(0)
            return

        chan = chan.replace(" ","")
        cmd = None
        for x in self.CMDS_DEFS["CV"].values:
            if x.name == chan:
                cmd = x.value
                break
        if cmd:
            fut = await self._send_command("CV",f"{cmd} DOWN")
            try:
                await self._wait_for_response_with_timeout(fut)
            except asyncio.TimeoutError:
                self._clean_fut(fut)
                raise AvrTimeoutError
        else:
            logging.error(f"Channel {chan} should exist. This should not have happened.")

    async def channels_bias_reset(self):
        fut = await self._send_command("CV", "ZRL")
        try:
            await self._wait_for_response_with_timeout(fut)
        except asyncio.TimeoutError:
            self._clean_fut(fut)
            raise AvrTimeoutError

    async def select_source(self, source: str) -> None:
        """Select the input source."""
        try:
            source = self.CMDS_DEFS["SI"].values[source.replace(" ","")]
        except:
            logging.warning(f"Warning: {source} is not a valid source")
            await asyncio.sleep(0)
            return
        fut = await self._send_command("SI", source.value)
        try:
            await self._wait_for_response_with_timeout(fut)
        except asyncio.TimeoutError:
            #Source might not be available
            logging.warning(f"Warning: {source} may not be a valid source for this device.")
            self._clean_fut(fut)

    async def select_sound_mode(self, mode: str) -> None:
        """Select the sound mode."""
        try:
            mode = self.CMDS_DEFS["MS"].values[mode.replace(" ","")]
        except:
            logging.warning(f"Warning: {mode} is not a valid mode")
            await asyncio.sleep(0)
            return
        fut = await self._send_command("MS", mode.value)
        try:
            await self._wait_for_response_with_timeout(fut)
        except:
            logging.warning(f"Warning: {mode} may not be a valid surround mode for this device.")
            self._clean_fut(fut)

    async def select_eco_mode(self, mode: str) -> None:
        """Select the sound mode."""
        try:
            mode = SurroundMode[mode]
        except:
            logging.warning(f"Warning: {mode} is not a valid eco  mode")
            await asyncio.sleep(0)
            return
        fut = await self._send_command("ECO", mode.value)
        await self._wait_for_response_with_timeout(fut)

    def notifyme(self, func: Callable) -> None:
        """Register a callback for when an event happens. The callable should have 2 parameters,
        The label of the the changing value and the new value
        """
        self.notify = func

    def close(self):
        self.alive = False
        self._writer.close()
        self.rtask.cancel()

    #API ends here

    async def _send_command(self, cmd: str, val: Any) -> asyncio.Future:
        loop = asyncio.get_event_loop()
        fut = loop.create_future()
        self.queue.append([cmd,fut])
        tosend=f"{cmd}{val}\r"
        logging.debug(f"Sending {tosend}")
        self._writer.write(tosend.encode())
        await self._writer.drain()
        return fut

    async def _with_timeout(self, fut) -> Optional[Any]:
        try:
            return await asyncio.wait_for(fut, self._timeout)
        except asyncio.TimeoutError:
            raise AvrTimeoutError

    async def _wait_for_response_with_timeout(self, fut: asyncio.Future) -> None:
        await self._with_timeout(self._wait_for_response(fut))

    async def _wait_for_response(self, fut: asyncio.Future) -> None:
        await fut

    def _process_response(self, response: str) -> Optional[str]:
        matches = [cmd for cmd in self.CMDS_DEFS.keys() if response.startswith(cmd)] + \
                    [cmd for cmd in EXTRAS if response.startswith(cmd)]

        if not matches:

            return None

        if len(matches) > 1:
            matches.sort(key=len, reverse=True)
        match = matches[0]

        if getattr(self,"_parse_"+match, None):
            getattr(self,"_parse_"+match)(response.strip()[len(match):])
        else:
            #A few special cases ... for now
            if response.startswith("SSINFAISFSV"):
                try:
                    sr = int(response.split(" ")[-1])
                    if sr > 200:
                        sr = round(sr/10,1)
                    else:
                        sr = float(sr)
                    self.status["Sampling Rate"] = sr
                except Exception as e:
                    logging.debug(f"Error with sampling rate: {e}")
            else:
                logging.warning(f"Warning _parse_{match} is not defined.")

        return match

    def _parse_many(self, cmd: str, resp: str) -> None:
        for x in self.CMDS_DEFS[cmd].values:
            if resp == x.value:
                lbl = self.CMDS_DEFS[cmd].label
                if self.status[lbl] != cc_string(x.name):
                    self.status[lbl] = cc_string(x.name)
                    if self.notify:
                        self.notify(lbl,self.status[lbl])

    def _parse_MV(self, resp: str) -> None:
        level = only_int(resp)

        if level:
            if len(level)>2:
                level=int(level)/10
            else:
                level=float(level)

            if resp.startswith("MAX"):
                self.maxvol = level
            else:
                lbl = self.CMDS_DEFS["MV"].label
                if self.status[lbl] != level:
                    self.status[lbl] = level
                    if self.notify:
                        self.notify(lbl,self.status[lbl])

    def _parse_PW(self, resp: str) -> None:
        self._parse_many("PW",resp)

    def _parse_ECO(self, resp: str) -> None:
        self._parse_many("ECO",resp)

    def _parse_SI(self, resp: str) -> None:
        self._parse_many("SI",resp)

    def _parse_MU(self, resp: str) -> None:
        nval = resp == "ON"
        lbl = self.CMDS_DEFS["MU"].label
        if self.status[lbl] != nval:
            self.status[lbl] = nval
            if self.notify:
                self.notify(lbl,self.status[lbl])

    def _parse_CV(self, resp: str) -> None:
        """ Different here... Needs to be reset"""
        if resp=="END":
            self.cvend = True
            if self.notify:
                lbl = self.CMDS_DEFS["CV"].label
                self.notify(lbl,self.status[lbl])
        else:
            if self.cvend:
                self.status[self.CMDS_DEFS["CV"].label] = {}
                self.cvend = False
            spkr, level = resp.split(" ")

            if level:
                if len(level)>2:
                    level=int(level)/10
                else:
                    level=float(level)
            level -= 50
            for x in self.CMDS_DEFS["CV"].values:
                if x.value == spkr:
                    spkrname = cc_string(x.name)
                    break
            try:
                self.status[self.CMDS_DEFS["CV"].label][spkrname] = level
            except:
                logging.debug(f"Unknown speaker code {spkr}")

    def _parse_MS(self, resp: str) -> None:
        """ Different here... What we get is not what we send. So we try to transform
        the result through semi-cllever string manipulation
        """

        resp = resp.replace("+", " ")
        resp = " ".join([x.title() for x in resp.split(" ")])
        for old,new in SSTRANSFORM:
            resp = resp.replace(old,new)
        #Clean up spaces
        resp = re.sub(r'[_\W]+', ' ', resp)
        lbl = self.CMDS_DEFS["MS"].label
        if self.status[lbl] != resp:
            self.status[lbl]=resp
            if self.notify:
                self.notify(lbl,self.status[lbl])

    def _clean_fut(self,fut):
        for idx in range(len(self.queue)):
            if self.queue[idx][1] == fut:
                self.queue[idx][1].cancel()
                del(self.queue[idx])
                break

    async def _do_read(self):
        """ Keep on reading the info coming from the AVR"""

        asyncio.get_event_loop().create_task(self.refresh())
        while self.alive:
            data =b''
            while not data or data[-1] != ord('\r'):
                if self.queue:
                    try:
                        data += await asyncio.wait_for(self._reader.read(1), self._timeout)
                    except asyncio.TimeoutError:
                        #Gone
                        self.alive = False
                        self._writer.close()
                        await self._writer.wait_close()
                        break
                else:
                    data += await self._reader.read(1)

            logging.debug(f"Received: {data}")
            match = self._process_response(data.decode().strip("\r"))

            for idx in range(len(self.queue)):
                cmd, fut = self.queue[idx]
                if cmd == match:
                    fut.set_result(True)
                    del(self.queue[idx])
                    break



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# This application is an asyncio library to discover HEOS devices
#
# Copyright (c) 2020 François Wautier
#
# Note large part of this code was taken from scapy and other opensource software
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE

import aiohttp
import asyncio
import logging
import socket
import struct
from xml.dom.minidom import parseString

MULTICAST_PORT = 1900
MULTICAST_ADDR = "239.255.255.250"
MULTICAST6_ADDR = "ff05::c"
#MULTICAST6_ADDR = "ff08::c"
#MULTICAST6_ADDR = "ff0e::c"

class DiscoveryClientProtocol:
    def __init__(self, loop,addr=MULTICAST_ADDR, callb=None,timeout=5):
        """Class representing a device discovery protocol

        :param loop: The asyncio loop
        :type name: asyncio.unix_events._UnixSelectorEventLoop
        :param addr: the broadcast address.
        :type addr: str
        :param callb: callback to register discovered instances
        :type callb: function
        :returns: protocol instance.
        :rtype: DiscoveryClientProtocol

    """
        """
        Here we
        """
        self.loop = loop
        self.transport = None
        self.addr=addr
        self.timeout=timeout
        self.callb = callb

    def connection_made(self, transport):
        """
        Once connected, broadcast the SSDP discovery message
        """

        self.transport = transport
        sock = self.transport.get_extra_info('socket')
        addrinfo = socket.getaddrinfo(self.addr, None)[0]
        ttl = struct.pack('@i', 1)
        if addrinfo[0] == socket.AF_INET: # IPv4
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)
        else:
            sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, ttl)
        request = '\r\n'.join(("M-SEARCH * HTTP/1.1",
                               "HOST:{}:{}",
                               "ST:upnp:rootdevice",
                               #"ST:ssdp:all",
                               "MX:2",
                               'MAN:"ssdp:discover"',
                               "", "")).format(self.addr,MULTICAST_PORT)
        self.transport.sendto(request.encode(), (self.addr,MULTICAST_PORT))
        self.loop.call_later(self.timeout,self.endme)


    def datagram_received(self, data, addr):
        """
            After receiving a response to our discovery broadcast, let's
            make surte this is a heos device.
        """
        if "denon-heos" in data.decode().lower():
            for x in data.decode("ascii").split("\r\n"):
                if x.upper().startswith("LOCATION:"):
                    loc=x.replace("LOCATION:","").strip()
            #We got a possible match. Let's investiate further
            self.loop.create_task(self.get_info(addr[0],loc))

    def error_received(self, exc):
        pass

    def connection_lost(self, exc):
        """
            Dead, we are
        """
        self.transport.close()

    def endme(self):
        """
            Ending discovery
        """
        self.transport.close()


    async def get_info(self,addr, url):
        """
            Let's analyse the data provided by the device
            answering our discovery call

                :param addr: The device IP address
                :type name: str
                :param url: The location URL provided by the device.
                :type url: str
                :returns: Nothing
                :rtype: None
        """

        def getText(nodelist):
            rc = []
            for node in nodelist:
                if node.nodeType == node.TEXT_NODE:
                    rc.append(node.data)
            return ''.join(rc)

        try:
            txt = None
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    txt = await response.text()
            if txt:
                data = parseString(txt)
                dev = data.getElementsByTagName("device")[0]
                rdata={"ip": addr}
                rdata["brand"] = getText(dev.getElementsByTagName("manufacturer")[0].childNodes)
                rdata["model"] = getText(dev.getElementsByTagName("modelName")[0].childNodes)
                rdata["serial"] = getText(dev.getElementsByTagName("serialNumber")[0].childNodes)
                rdata["name"] = getText(dev.getElementsByTagName("friendlyName")[0].childNodes)
                logging.debug(f"Got device: {rdata}")
                if self.callb:
                    self.callb(rdata)
        except Exception as e:
            logging.error(f"Error: Error when parsing location XML: {e}")


async def start_discovery( addr=MULTICAST_ADDR, callb=None):
    loop = asyncio.get_event_loop()
    addrinfo = socket.getaddrinfo(MULTICAST_ADDR, None)[0]
    sock = socket.socket(addrinfo[0], socket.SOCK_DGRAM)
    connect = loop.create_datagram_endpoint(
        lambda: DiscoveryClientProtocol(loop,MULTICAST_ADDR, callb=callb),
        sock=sock
        )

    await connect

if __name__ == "__main__":
    import asyncio as aio

    def cb(data):
        print("Got device {}".format(data))

    loop = aio.get_event_loop()
    loop.run_until_complete(start_discovery(MULTICAST_ADDR,cb) )
    try:
        loop.run_until_complete(aio.sleep(6))
    except:
        pass
    finally:
        loop.close()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# This application is an example on how to use aiomadeavr
#
# Copyright (c) 2020 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
import sys
import asyncio as aio
import aiomadeavr as avr
from functools import partial
import argparse
import logging

def quick_print(val,depth=1):
    if isinstance(val,dict):
        for k,v in val.items():
            if isinstance(v,dict):
                print("\t"*depth + k+": ")
                quick_print(v,depth+1)
            else:
                print("\t"*depth + f"{k}:\t{v}")

#Simple device control from console
class devices():
    """ A simple class with a register and  unregister methods
    """
    def __init__(self, debug):
        self.devices={}
        self.doi=None #device of interest
        self.secondary = None  #Either "source", "surround", or channel
        self.debug = debug

    def register(self,info):
        if "serial" in info and info["serial"].lower() not in self.devices:
            aio.create_task(self.set_device(info))
        else:
            if not self.devices[info["serial"].lower()].alive:
                aio.create_task(self.set_device(info))
    def unregister(self,mac):
        if mac.lower() in self.devices:
            logging.debug("%s is gone"% self.devices[mac.lower()].name)
            self.devices[mac.lower()].stop()
            del(self.devices[mac.lower()])

    async def set_device(self, info):
        logging.debug(f"Adding {info}")
        try:
            newdev = await avr.avr_factory(info["name"],info["ip"])
            if newdev:
                self.devices[info["serial"].lower()] = newdev
                if self.debug:
                    self.devices[info["serial"].lower()].notifyme(notification)
        except:
            logging.warning(f"Could not connect to {info['ip']}.")

    def stop(self):
        for dev in self.devices.values():
            dev.close()


def readin():
    """Reading from stdin and displaying menu"""

    selection = sys.stdin.readline().strip("\n")

    loaddr = [ x for x in MyDevices.devices.keys()]
    loaddr.sort()
    lov=[ x for x in selection.split(" ") if x != ""]
    if lov:
        if MyDevices.doi:
            if MyDevices.secondary:
                if MyDevices.secondary == "source":
                    loi =  MyDevices.doi.source_list
                    loi.sort()
                    try:
                        sel = int(lov[0])-1
                        if sel >= 0 and sel<len(loi):
                            source = loi[sel]
                            aio.create_task(MyDevices.doi.select_source(source))
                        else:
                            print("\nError: Source selection incorrect.\n")
                    except:
                        print("\nError: Source selection must be a number.\n")
                elif MyDevices.secondary == "surround":
                    loi =  MyDevices.doi.sound_mode_list
                    loi.sort()
                    try:
                        sel = int(lov[0])-1
                        if sel >= 0 and sel<len(loi):
                            mode = loi[sel]
                            aio.create_task(MyDevices.doi.select_sound_mode(mode))
                        else:
                            print("\nError: Surround Mode  selection incorrect.\n")
                    except:
                        print("\nError: Surround Mode selection must be a number.\n")
                elif MyDevices.secondary == "channel":
                    loi =  MyDevices.doi.channels_bias_list
                    loi.sort()
                    try:
                        sel = int(lov[0])-1
                        if sel == len(loi):
                            aio.create_task(MyDevices.doi.channels_bias_reset())
                        elif sel >= 0 and sel<len(loi):
                            if lov[1].lower() == "up":
                                aio.create_task(MyDevices.doi.channel_bias_up(loi[sel]))
                            elif lov[1].lower() == "down":
                                aio.create_task(MyDevices.doi.channel_bias_down(loi[sel]))
                            else:
                                level = float(lov[1])
                                aio.create_task(MyDevices.doi.set_channel_bias(loi[sel],level))
                    except:
                        print("\nError: Channel Bias meeds a channel and a level (float, up or down).\n")

                MyDevices.doi=None
                MyDevices.secondary = None

            else:
                #try:
                if int(lov[0]) == 0:
                    MyDevices.doi=None
                elif int(lov[0]) == 1:
                    if len(lov) >1:
                        if lov[1].lower() in ["1","on","true"]:
                            aio.create_task(MyDevices.doi.turn_on())
                        else:
                            aio.create_task(MyDevices.doi.turn_off())
                        MyDevices.doi=None
                    else:
                        print("Error: For power you must indicate on or off\n")
                elif int(lov[0]) == 2:
                    aio.create_task(MyDevices.doi.mute_volume(not MyDevices.doi.is_volume_muted))
                    MyDevices.doi=None
                elif int(lov[0]) == 3:
                    if len(lov) >1:
                        if lov[1].lower() == "up":
                            aio.create_task(MyDevices.doi.volume_up())
                        elif lov[1].lower() == "down":
                            aio.create_task(MyDevices.doi.volume_down())
                        else:
                            try:
                                lvl = float(lov[1])
                                aio.create_task(MyDevices.doi.set_volume(lvl))
                            except:
                                print(f"Error: {lov[1]} is not a float.")
                                print("Error: For volume you must specify 'up', 'down' or a float value.\n")
                    else:
                        print("Error: For volume you must specify 'up', 'down' or a float value.\n")
                    MyDevices.doi=None

                elif int(lov[0]) == 4:
                    MyDevices.secondary = "source"
                    los = MyDevices.doi.source_list
                    los.sort()
                    print("Select source for {}:".format(MyDevices.doi.name))
                    idx = 1
                    for src in los:
                        print(f"\t[{idx}]\t{src}")
                        idx += 1
                elif int(lov[0]) == 5:
                    MyDevices.secondary = "surround"
                    los = MyDevices.doi.sound_mode_list
                    los.sort()
                    print("Select sound mode for {}:".format(MyDevices.doi.name))
                    idx = 1
                    for src in los:
                        print(f"\t[{idx}]\t{src}")
                        idx += 1
                elif int(lov[0]) == 6:
                    MyDevices.secondary = "channel"
                    los = MyDevices.doi.channels_bias_list
                    los.sort()
                    print("Select channel bias mode for {}:".format(MyDevices.doi.name))
                    idx = 1
                    for src in los:
                        print(f"\t[{idx}]\t{src} <level>|up|down")
                        idx += 1
                    print(f"\t[{idx}]\tReset all channels")

                elif int(lov[0]) == 7:
                    print(f"Status for {MyDevices.doi.name}")
                    quick_print(MyDevices.doi.status)
                    MyDevices.doi=None
                #except:
                    #print ("\nError: Selection must be a number.\n")
        else:
            try:
                if int(lov[0]) > 0:
                    if int(lov[0]) <=len(MyDevices.devices):
                        MyDevices.doi=MyDevices.devices[loaddr[int(lov[0])-1]]
                    else:
                        print("\nError: Not a valid selection.\n")

            except:
                print ("\nError: Selection must be a number.\n")


    if MyDevices.doi:
        if MyDevices.secondary is None:
            print("Select Function for {}:".format(MyDevices.doi.name))
            print("\t[1]\tPower (0 or 1)")
            print("\t[2]\tToggle Mute")
            print("\t[3]\tVolume [float val|up|down]")
            print("\t[4]\tSource")
            print("\t[5]\tSurround Mode")
            print("\t[6]\tChannel Bias")
            print("\t[7]\tInfo")
            print("")
            print("\t[0]\tBack to device selection")
    else:
        idx=1
        print("Select Device:")
        for x in loaddr:
            print("\t[{}]\t{}".format(idx,MyDevices.devices[x].name or x))
            idx+=1
    print("")
    print("Your choice: ", end='',flush=True)

def notification(lbl, value):
    print(f"Value for {lbl} changed to {value}")

parser = argparse.ArgumentParser(description="Track and interact with Marantz/Denon devices.")
parser.add_argument("-v","--verbose", default=False, action="store_true",
                    help="Print more information.")
parser.add_argument("-d","--debug", default=False, action="store_true",
                    help="Print debug information.")

try:
    opts = parser.parse_args()
except Exception as e:
    parser.error("Error: " + str(e))

if opts.debug:
    logging.basicConfig(level=logging.DEBUG)
elif opts.verbose:
    logging.basicConfig(level=logging.INFO)
else:
    logging.basicConfig(level=logging.ERROR)


MyDevices= devices(opts.debug)
loop = aio.get_event_loop()
loop.run_until_complete(avr.start_discovery(callb=MyDevices.register))
try:
    loop.add_reader(sys.stdin,readin)
    print("Hit \"Enter\" to start")
    print("Use Ctrl-C to quit")
    loop.run_forever()
except:
    print("Exiting at user's request.")
finally:
    MyDevices.stop()
    loop.remove_reader(sys.stdin)
    loop.run_until_complete(aio.sleep(2))
    loop.close()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# Originaly from https://github.com/silvester747/aio_marantz_avr
#
# These are messages one can receive from the device.
#
# Copyright (c) 2020 François Wautier
#
# Note large part of this code was taken from scapy and other opensource software
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE

"""Enums used by the AVR."""

from enum import Enum


class Power(Enum):
    Off = "OFF"
    On = "ON"
    Standby = "STANDBY"

class ChannelBias(Enum):
    FrontLeft = "FL"
    FrontRight = "FR"
    Centre = "C"
    Subwoofer = "SW"
    Subwoofer2 = "SW2"
    SurroundLeft = "SL"
    SurroundRight = "SR"
    SurroundBackLeft = "SBL"
    SurroundBackRight = "SBR"
    FrontHeightLeft = "FHL"
    FrontHeightRight = "FHR"
    FrontWideLeft = "FWL"
    FronWideRight = "FWR"
    FrontTopLeft = "TFL"
    FrontTopRight = "TFR"
    MiddleTopLeft = "TML"
    MiddleTopRight = "TMR"
    RearTopLeft = "TRL"
    RearTopRight = "TRR"
    RearHeightLeft = "RHL"
    RearHeightRight = "RHR"
    FrontDolbyLeft = "FDL"
    FrontDolbyRight = "FDR"
    SurroundDolbyLeft = "SDL"
    SurroundDolbyRight = "SDR"
    BackDolbyLeft = "BDL"
    BackDolbyRight = "BDR"
    SurroundHeightLeft = "SHL"
    SurroundHeightRight = "SHR"
    TopSurround = "TS"
    CentreHeight = "CH"

class InputSource(Enum):
    Phono = "PHONO"
    CD = "CD"
    DVD = "DVD"
    Bluray = "BD"
    TV = "TV"
    SetTopBox = "SAT/CBL"
    MediaPlayer = "MPLAY"
    Game = "GAME"
    Tuner = "TUNER"
    HDRadio = "HDRADIO"
    SiriusXM = "SIRIUSXM"
    Pandora = "PANDORA"
    InternetRadio = "IRADIO"
    Server = "SERVER"
    Favourites = "FAVORITES"
    Aux1 = "AUX1"
    Aux2 = "AUX2"
    Aux3 = "AUX3"
    Aux4 = "AUX4"
    Aux5 = "AUX5"
    Aux6 = "AUX6"
    Aux7 = "AUX7"
    OnlineMusic = "NET"
    Bluetooth = "BT"

class AudioInput(Enum):
    Auto = "AUTO"
    HDMI = "HDMI"
    Digital = "DIGITAL"
    Analog = "ANALOG"
    MultiChannel = "7.1IN"
    NoSound = "NO"

class EcoMode(Enum):
    Off = "OFF"
    On = "ON"
    Auto = "AUTO"

class SurroundMode(Enum):
    # Settable values
    Movie = "MOVIE"
    Music = "MUSIC"
    Game = "GAME"
    Direct = "DIRECT"
    PureDirect = "PURE DIRECT"
    Stereo = "STEREO"
    Auto = "AUTO"
    DolbyDigital = "DOLBY DIGITAL"
    DtsSurround = "DTS SURROUND"
    Auro3D = "AURO3D"
    Auro2DSurround = "AURO2DSURR"
    MultiChannelStereo = "MCH STEREO"
    Virtual = "VIRTUAL"

    # Rotate between options
    Left = "LEFT"
    Right = "RIGHT"




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
import setuptools
version = '0.1'

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(name='aiomadeavr',
    packages=['aiomadeavr'],
    #packages=setuptools.find_packages(),
    version=version,
    author='François Wautier',
    author_email='francois@wautier.eu',
    description='Library/utility to control Marantz/Denon devices over telnet.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/frawau/aiomadeavr',
    keywords = ['Marantz', 'Denon', 'Control', "Network"],
    license='MIT',
    install_requires=[
    "aiohttp"
    ],
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6'
    ],
    zip_safe=False)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# Exposed API for aiomadeavr
#
# Copyright (c) 2020 François Wautier
#
# Note large part of this code was taken from scapy and other opensource software
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
from .avr import avr_factory, AvrError, AvrTimeoutError
from .discovery import start_discovery
__version__ = '0.1'
